import os

workers_location = os.path.dirname(__file__)


# ----------------------------------------------------------------------
def select_worker(worker):
    """"""
    return os.path.join(workers_location, worker)


# ----------------------------------------------------------------------
def list_workers():
    """"""
    workers = []
    for item in os.listdir(workers_location):
        if os.path.isdir(os.path.join(workers_location, item)) and not item.startswith('__'):
            workers.append(item)
    return workers




import docker
import logging


########################################################################
class Swarm:
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, base_url='unix://var/run/docker.sock', advertise_addr=''):
        """Constructor"""
        self.client = docker.DockerClient(base_url=base_url)

        try:
            logging.warning("Starting swarm...")
            if not self.client.swarm.attrs.get('ID'):
                swarm_init_response = self.client.swarm.init(advertise_addr=advertise_addr)
                logging.warning(f"Swarm started, ID: {swarm_init_response}")
            else:
                logging.warning(f"Swarm is already running, ID: {self.client.swarm.attrs.get('ID')}")
        except docker.errors.APIError as e:
            logging.warning("Error:", str(e))
            return

        self.create_networks()

    # ----------------------------------------------------------------------
    def create_networks(self):
        """"""
        self.networks = ['hci_network']
        logging.warning("Creating networks...")
        for network in self.networks:
            if network not in [n.name for n in self.client.networks.list()]:
                self.client.networks.create(network, driver="overlay")
                logging.warning(f"Created network '{network}'")

    # ----------------------------------------------------------------------
    def create_volume(self, volume_name):
        """"""
        logging.warning("Creating volumes...")
        if not volume_name in self.volumes:
            if not volume_name.endswith('-volume'):
                volume_name = f'{volume_name}-volume'
            self.client.volumes.create(name=volume_name)
        else:
            logging.warning(f"Volume '{volume_name}' already exists")

        return volume_name

    # ----------------------------------------------------------------------
    @property
    def services(self, attr='name'):
        """"""
        return [getattr(service, attr) for service in self.client.services.list()]

    # # ----------------------------------------------------------------------
    # def desribe_service(self, service, attr='name'):
        # """"""
        # return [getattr(service, attr) for service in self.client.services.list()]

    # ----------------------------------------------------------------------
    @property
    def containers(self, attr='id'):
        """"""
        return [getattr(container, attr) for container in self.client.containers.list()]

    # # ----------------------------------------------------------------------
    # def containers_get(self, attr='id'):
        # """"""
        # return [getattr(container, attr) for container in self.client.containers.list()]

    # ----------------------------------------------------------------------
    @property
    def volumes(self):
        """"""
        return [v.name for v in self.client.volumes.list() if v.name.endswith('-volume')]

    # ----------------------------------------------------------------------
    def stop_service(self, service_name):
        """"""
        service = self.client.services.get(service_name)
        service.remove()

    # ----------------------------------------------------------------------
    def stop_all_services(self):
        """"""
        for service in self.services:
            self.stop_service(service)

    # ----------------------------------------------------------------------
    def stats(self, service_name):
        """"""
        service = self.client.services.get(service_name)
        stats = []
        for task in service.tasks():
            container_id = task['Status']['ContainerStatus']['ContainerID']
            if container_id in self.containers:
                stats.append(self.client.containers.get(container_id).stats(stream=False))
        return stats

    # ----------------------------------------------------------------------
    def start_jupyter(self, service_name="jupyterlab-service", port=8888, restart=False):
        """"""
        if restart and (service_name in self.services):
            self.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        volume_name = self.create_volume(service_name)
        service = self.client.services.create(
            image="dunderlab/python311:latest",
            name=service_name,
            networks=self.networks,
            command=["jupyter", "lab", "--notebook-dir='/app'",
                     "--ip=0.0.0.0", "--port=8888",
                     "--allow-root", "--NotebookApp.token=''", "--NotebookApp.password=''"],
            endpoint_spec=docker.types.EndpointSpec(ports={8888: port}),
            mounts=[
                docker.types.Mount(
                    type='bind',
                    source='/var/run/docker.sock',
                    target='/var/run/docker.sock'
                ),
                docker.types.Mount(
                    target='/app',
                    source=volume_name,
                    type="volume",
                    read_only=False
                ),
            ]
        )
        return service_name in self.services

    # ----------------------------------------------------------------------
    def start_kafka(self, kafka_service_name="kafka-service",
                    zookeeper_service_name="zookeeper-service",
                    kafka_port=9092, kafka_port_external=19092,
                    zookeeper_port=2181, restart=False):
        """"""
        if restart and (kafka_service_name in self.services):
            self.stop_service(kafka_service_name)
            logging.warning(f"Restarting service '{kafka_service_name}'")

        if restart and (zookeeper_service_name in self.services):
            self.stop_service(zookeeper_service_name)
            logging.warning(f"Restarting service '{zookeeper_service_name}'")

        if not kafka_service_name in self.services:
            kafka_service = self.client.services.create(
                image="dunderlab/kafka:latest",
                # restart_policy=docker.types.RestartPolicy(condition='any'),
                name=kafka_service_name,
                networks=self.networks,
                endpoint_spec=docker.types.EndpointSpec(ports={kafka_port: kafka_port,
                                                               kafka_port_external: kafka_port_external, }),
                env=[
                    f"KAFKA_ZOOKEEPER_CONNECT={zookeeper_service_name}:{zookeeper_port}",
                    f"KAFKA_LISTENER_SECURITY_PROTOCOL_MAP=PLAINTEXT:PLAINTEXT,PLAINTEXT_HOST:PLAINTEXT",
                    f"KAFKA_ADVERTISED_LISTENERS=PLAINTEXT://{kafka_service_name}:{kafka_port},PLAINTEXT_HOST://localhost:{kafka_port_external}",
                ],
            )
        else:
            logging.warning(f"Service '{kafka_service_name}' already exist")

        if not zookeeper_service_name in self.services:
            zookeeper_service = self.client.services.create(
                image="dunderlab/zookeeper:latest",
                name=zookeeper_service_name,
                networks=self.networks,
                endpoint_spec=docker.types.EndpointSpec(ports={zookeeper_port: zookeeper_port}),
                env=[
                    f"ZOOKEEPER_CLIENT_PORT={zookeeper_port}",
                ],
            )
        else:
            logging.warning(
                f"Service '{zookeeper_service_name}' already exist")

        return kafka_service_name in self.services, zookeeper_service_name in self.services

    # ----------------------------------------------------------------------
    def start_kafka_logs(self, kafka_service_name="kafka-logs-service",
                         zookeeper_service_name="zookeeper-logs-service",
                         kafka_port=9093, kafka_port_external=19093,
                         zookeeper_port=2182, restart=False):
        """"""
        return self.start_kafka(kafka_service_name, zookeeper_service_name, kafka_port, kafka_port_external, zookeeper_port, restart)

    # ----------------------------------------------------------------------
    def start_timescaledb(self, service_name="timescaledb-service", port=5432, volume_name=None, restart=False):
        """"""
        if restart and (service_name in self.services):
            self.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        if volume_name is None:
            volume_name = self.create_volume(service_name)
        else:
            volume_name = self.create_volume(volume_name)

        timescaledb_service = self.client.services.create(
            image="timescale/timescaledb:latest-pg15",
            name=service_name,
            networks=self.networks,
            env=[
                "POSTGRES_PASSWORD=password",
                "POSTGRES_USER=postgres",
                "POSTGRES_DB=timescaledb",
                "POSTGRES_MAX_CONNECTIONS=500"
            ],
            endpoint_spec=docker.types.EndpointSpec(ports={5432: port}),
            mounts=[
                docker.types.Mount(
                    target='/var/lib/postgresql/data',
                    source=volume_name,
                    type="volume",
                    read_only=False
                ),
            ]
        )
        return service_name in self.services

    # ----------------------------------------------------------------------
    def get_join_command(self):
        """"""
        swarm_info = self.client.info().get('Swarm')
        if swarm_info and swarm_info.get('ControlAvailable'):
            worker_join_token = self.client.swarm.attrs['JoinTokens']['Worker']
            manager_addr = swarm_info.get('RemoteManagers')[0].get('Addr')
            return f'docker swarm join --token {worker_join_token} {manager_addr}'

    # ----------------------------------------------------------------------
    def advertise_addr(self):
        """"""
        swarm_info = self.client.info().get('Swarm')
        if swarm_info and swarm_info.get('ControlAvailable'):
            # worker_join_token = self.client.swarm.attrs['JoinTokens']['Worker']
            manager_addr = swarm_info.get('RemoteManagers')[0].get('Addr')
            return manager_addr









from radiant.framework.server import PythonHandler
from foundation.utils import kafkalogs
import logging as logging_orig


########################################################################
class logging(PythonHandler):
    """"""

    # # ----------------------------------------------------------------------
    # def __getattr__(self, attribute):
        # """"""
        # instance = getattr(logging_orig, attribute)
        # if callable(instance):
            # def inset(*args, **kwargs):
                # """"""
                # instance(*args, **kwargs)
            # return inset
        # else:
            # return instance

    # ----------------------------------------------------------------------
    def warning(self, msg):
        """"""
        logging_orig.warning(msg)




from radiant.framework.server import RadiantAPI, RadiantServer, render
from radiant.framework.utils import environ, run_script

import os


FigureStream = None


########################################################################
class FrameworkAPI(RadiantAPI):
    """Rename Randiant with a arand new class."""

    # ---------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # ----------------------------------------------------------------------
    def __new__(self, *args, **kwargs):
        """"""
        FrameworkServer(self.__name__, *args, **kwargs)


# ----------------------------------------------------------------------
def FrameworkServer(class_, swarm=True, logs=True, *args, **kwargs):
    """Rename `RadiantServer` with a preconfigured `StimuliServer`."""

    # brython_environ = {k: os.environ[k] for k in os.environ if k.startswith('BCISTREAM_')}
    # brython_environ = {
        # k: os.environ.get(k)
        # for k in dict(os.environ)
        # if k.startswith('BCISTREAM_')
    # }
    brython_environ = {
        'STREAM': environ('STREAM', '5001'),
        'RADIANT': environ('RADIANT', '5002'),
    }

    if scripts := kwargs.get('scripts'):
        for script, port in scripts:
            if environ(port):
                port = environ(port)
            run_script(script, port)

    python_tools = []
    if swarm:
        python_tools.append([os.path.join(os.path.dirname(os.path.abspath(__file__)), 'python_tools',
                                          'python_swarm.py'), 'swarm', '/python_swarm'])
    if logs:
        python_tools.append([os.path.join(os.path.dirname(os.path.abspath(__file__)), 'python_tools',
                                          'python_logger.py'), 'logging', '/python_logger'],)

    if 'python' in kwargs:
        kwargs['python'] = list(kwargs['python']) + python_tools
    else:
        kwargs['python'] = python_tools

    return RadiantServer(
        class_,
        path=os.path.realpath(os.path.join(os.path.dirname(__file__), 'brython')),
        # handlers=(
        # [
        # r'^/ws',
        # (
        # os.path.realpath(
        # os.path.join(
        # os.path.dirname(__file__), 'tornado_handlers.py'
        # )
        # ),
        # 'WSHandler',
        # ),
        # {},
        # ],
        # [
        # r'^/dashboard',
        # RadiantHandler,
        # {
        # 'mode': 'dashboard',
        # },
        # ],
        # [
        # r'^/mode',
        # (
        # os.path.realpath(
        # os.path.join(
        # os.path.dirname(__file__), 'tornado_handlers.py'
        # )
        # ),
        # 'ModeHandler',
        # ),
        # {},
        # ],
        # ),

        template=os.path.realpath(
            os.path.join(os.path.dirname(__file__), 'template.html')
        ),

        modules=[
            'md_git',
            'material_symbols',
            'material_icons',
        ],

        environ=brython_environ,
        port=environ('RADIANT', '5002'),
        host='0.0.0.0',

        # theme=os.path.realpath(
        # os.path.join(os.path.dirname(__file__), 'colors.xml')
        # ),

        # callbacks=[(os.path.realpath(os.path.join(
        # os.path.dirname(__file__), 'tornado_handlers.py')), 'consumer')]
        debug_level=0,
        brython_version='3.11.1',
        **kwargs,
    )




import os
import random
import socket
import logging

from string import digits, ascii_letters
import docker
from foundation.utils.swarm import Swarm
from foundation.workers import select_worker

WORKER_NAME = "{}-service-worker"


########################################################################
class Workers:
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, swarm=None, swarm_advertise_addr=None):
        """Constructor"""

        if (swarm_advertise_addr) and (swarm is None):
            self.swarm = Swarm(advertise_addr=swarm_advertise_addr)
        elif (swarm_advertise_addr is None) and (swarm):
            self.swarm = swarm
        else:
            self.swarm = Swarm()

    # ----------------------------------------------------------------------
    def gen_worker_name(self, length=8):
        """"""
        id_ = "".join([random.choice(ascii_letters + digits) for _ in range(length)])
        if not WORKER_NAME.format(id_) in self.swarm.services:
            return WORKER_NAME.format(id_)
        return self.gen_worker_name(length)

    # ----------------------------------------------------------------------
    def get_open_port(self,):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    # ----------------------------------------------------------------------
    def stop_all_workers(self):
        """"""
        for worker in self.swarm.services:
            if worker.endswith(WORKER_NAME.format("")):
                self.swarm.stop_service(worker)

    # ----------------------------------------------------------------------
    def start_django_worker(self, worker_path, service_name=None, port=None, restart=False, version='1.1', endpoint=''):
        """"""
        if os.path.isabs(worker_path) or os.path.exists(worker_path):
            worker_path = os.path.abspath(worker_path)
            app_name = os.path.split(worker_path)[-1]
        elif os.path.exists(select_worker(worker_path)):
            app_name = worker_path
            service_name = WORKER_NAME.format(worker_path.replace("_", "-"))
            worker_path = select_worker(worker_path)
        else:
            logging.warning(f"Service {worker_path} doesn't exits.")
            return

        if port is None:
            port = self.get_open_port()

        if service_name is None:
            service_name = self.gen_worker_name()
        else:
            service_name = WORKER_NAME.format(service_name)
        service_name_env = os.path.split(worker_path)[-1]

        if restart and (service_name in self.swarm.services):
            self.swarm.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.swarm.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        service = self.swarm.client.services.create(
            image=f"dunderlab/djangoship:{version}",
            name=service_name,
            networks=self.swarm.networks,
            endpoint_spec=docker.types.EndpointSpec(ports={port: 80}),
            mounts=[
                docker.types.Mount(
                    type="bind",
                    source=worker_path,
                    target="/app/djangoship",
                    read_only=False,
                ),
                docker.types.Mount(
                    type="bind",
                    source="/var/run/docker.sock",
                    target="/var/run/docker.sock",
                ),
            ],
            env={
                "DJANGOPROJECT": app_name,
                "PORT": port,
                "ENDPOINT": endpoint,
                "SERVICE_NAME": service_name_env,
            },
        )

        return port

    # ----------------------------------------------------------------------
    def start_brython_worker(self, worker_path, service_name=None, port_stream=None, port_radiant=None, run="main.py", restart=False, version='1.2'):
        """"""
        if os.path.isabs(worker_path) or os.path.exists(worker_path):
            worker_path = os.path.abspath(worker_path)
        elif os.path.exists(select_worker(worker_path)):
            service_name = WORKER_NAME.format(worker_path.replace("_", "-"))
            worker_path = select_worker(worker_path)

        if port_stream is None:
            port_stream = self.get_open_port()
        if port_radiant is None:
            port_radiant = self.get_open_port()

        if service_name is None:
            service_name = self.gen_worker_name()
        else:
            service_name = WORKER_NAME.format(service_name)
        service_name_env = os.path.split(worker_path)[-1]

        if restart and (service_name in self.swarm.services):
            self.swarm.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.swarm.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        service = self.swarm.client.services.create(
            image=f"dunderlab/python311:{version}",
            name=service_name,
            networks=self.swarm.networks,
            command=[
                "/bin/bash", "-c",
                f"if [ -f \"/app/worker/requirements.txt\" ]; then pip install --root-user-action=ignore -r /app/worker/requirements.txt; fi && python /app/worker/{run}",
            ],
            endpoint_spec=docker.types.EndpointSpec(
                ports={
                    port_radiant: port_radiant,
                    port_stream: port_stream,
                }
            ),
            mounts=[
                docker.types.Mount(
                    type="bind",
                    source=worker_path,
                    target="/app/worker",
                    read_only=False,
                ),
                docker.types.Mount(
                    type="bind",
                    source="/var/run/docker.sock",
                    target="/var/run/docker.sock",
                ),
            ],
            env={
                "STREAM": port_stream,
                "RADIANT": port_radiant,
                "PORT": port_radiant,
                "SERVICE_NAME": service_name_env,
            },
        )

        return port_radiant

    # ----------------------------------------------------------------------
    def start_python_worker(self, worker_path, service_name=None, port=None, run="main.py", restart=False, version='1.2'):
        """"""
        if os.path.isabs(worker_path) or os.path.exists(worker_path):
            worker_path = os.path.abspath(worker_path)
        elif os.path.exists(select_worker(worker_path)):
            service_name = WORKER_NAME.format(worker_path.replace("_", "-"))
            worker_path = select_worker(worker_path)

        if port is None:
            port = self.get_open_port()

        if service_name is None:
            service_name = self.gen_worker_name()
        else:
            service_name = WORKER_NAME.format(service_name)
        service_name_env = os.path.split(worker_path)[-1]

        if restart and (service_name in self.swarm.services):
            self.swarm.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.swarm.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        service = self.swarm.client.services.create(
            image=f"dunderlab/python311:{version}",
            name=service_name,
            networks=self.swarm.networks,
            command=[
                "/bin/bash", "-c",
                f"if [ -f \"/app/worker/requirements.txt\" ]; then pip install --root-user-action=ignore -r /app/worker/requirements.txt; fi && python /app/worker/{run}",
            ],
            endpoint_spec=docker.types.EndpointSpec(
                ports={
                    port: port,
                }
            ),
            mounts=[
                docker.types.Mount(
                    type="bind",
                    source=worker_path,
                    target="/app/worker",
                    read_only=False,
                ),
                docker.types.Mount(
                    type="bind",
                    source="/var/run/docker.sock",
                    target="/var/run/docker.sock",
                ),
            ],
            env={
                "PORT": port,
                "SERVICE_NAME": service_name_env,
            },
        )

        return port

    # ----------------------------------------------------------------------
    def start_worker(self, worker_path, **kwargs):
        """"""
        if os.path.isabs(worker_path) or os.path.exists(worker_path):
            worker_path = os.path.abspath(worker_path)
        elif os.path.exists(select_worker(worker_path)):
            worker_path = select_worker(worker_path)

        if os.path.exists(os.path.join(worker_path, 'manage.py')):
            logging.warning('Running a Django worker')
            return self.start_django_worker(worker_path, **kwargs)
        elif os.path.exists(os.path.join(worker_path, 'main.py')):
            with open(os.path.join(worker_path, 'main.py'), 'r') as file:
                content = file.read()
                if 'from hci_framework.radiant.server' in content:
                    logging.warning('Running a Brython worker')
                    return self.start_brython_worker(worker_path, **kwargs)
                else:
                    logging.warning('Running a Pyhton worker')
                    return self.start_python_worker(worker_path, **kwargs)
        else:
            logging.warning('Impossible to detect a Worker')








# TODO: Signal test enable by default !!!!


import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))
version_str = '0.1a12'

setup(
    name='dunderlab-foundation',
    version=version_str,
    packages=['foundation'],

    author='Yeison Cardona',
    author_email='yencardonaal@unal.edu.co',
    maintainer='Yeison Cardona',
    maintainer_email='yencardonaal@unal.edu.co',

    download_url='https://github.com/dunderlab/python-dunerlab.foundation',

    install_requires=[
        'docker',
        'radiant-framework',
    ],

    include_package_data=True,
    license='BSD-2-Clause',
    description="",

    long_description=README,
    long_description_content_type='text/markdown',

    python_requires='>=3.11',

    # entry_points={
    # 'console_scripts': ['bci-framework=bci_framework.__main__:main'],
    # },

    scripts=[
        "cmd/foundation_status",
        "cmd/foundation_logs",
        "cmd/foundation_start",
        "cmd/foundation_worker",
    ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Hardware :: Hardware Drivers',
    ],

)






from foundation.radiant.utils import environ
from figurestream import FigureStream
import time
import numpy as np


########################################################################
class Stream(FigureStream):
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.axis = self.add_subplot(111)
        self.x = np.array(range(100))
        self.t0 = time.time()
        self.y = [self.t0] * 100

        self.axis.set_title('FigureStream')
        self.axis.set_xlabel('Time [s]')
        self.axis.set_ylabel('Amplitude')

        self.axis.set_ylim(0, 60)
        self.line1, *_ = self.axis.plot(self.x, np.zeros(self.x.size))
        self.stream()

    # ----------------------------------------------------------------------
    def stream(self):
        """"""
        while True:
            self.y.pop(0)
            self.y.append(1 / (time.time() - self.t0))
            self.t0 = time.time()

            self.line1.set_ydata(self.y)
            self.feed()


if __name__ == '__main__':
    Stream(host='0.0.0.0', port=environ('PORT', '5001'))
